from decimal import Decimal

from chellow.e.aahedc import hh
from chellow.utils import ct_datetime, to_utc


def test_hh(sess, mocker):
    mock_ss = mocker.Mock()
    mock_ss.caches = {}
    datum = {"gsp-kwh": 1, "start-date": to_utc(ct_datetime(2020, 1, 1))}
    mock_ss.hh_data = [datum]
    rate = Decimal("1.0")
    aahedc_rate_script = {"aahedc_gbp_per_gsp_kwh": rate}
    mock_ss.non_core_rate.return_value = aahedc_rate_script
    hh(mock_ss)

    assert isinstance(datum["aahedc-rate"], float)



import json
from decimal import Decimal
from io import BytesIO

from sqlalchemy import select

from chellow.gas.engine import g_rates
from chellow.models import GContract, GRateScript
from chellow.rate_server import download
from chellow.utils import ct_datetime, date_format, to_utc

DAILY_THRESHOLD = 145


def vb(ds):
    for hh in ds.hh_data:
        rate = float(
            g_rates(ds.sess, ds.caches, "ccl", hh["start_date"], True)[
                "ccl_gbp_per_kwh"
            ]
        )

        hh["ccl"] = rate


def rate_server_import(sess, log, set_progress, paths):
    log("Starting to check for new gas CCL rates")
    g_contract_name = "ccl"
    g_contract = GContract.find_industry_by_name(sess, g_contract_name)
    if g_contract is None:
        g_contract = GContract.insert_industry(
            sess,
            g_contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1996, 4, 1)),
            None,
            {},
        )
        sess.commit()
    g_contract_props = g_contract.make_properties()

    if not g_contract_props.get("enabled", False):
        log(
            "The automatic importer is disabled. To enable it, edit "
            "the contract properties to set 'enabled' to true."
        )
        return

    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year, utility, rate_type, file_name = path
            if utility == "gas" and rate_type == "ccl":
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    for year, year_files in sorted(year_entries.items()):
        year_start = to_utc(ct_datetime(year, 4, 1))
        if year_start < g_contract.start_g_rate_script.start_date:
            continue
        rs = sess.execute(
            select(GRateScript).where(
                GRateScript.g_contract == g_contract,
                GRateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = g_contract.insert_g_rate_script(sess, year_start, {})

        if len(year_files) > 0:
            file_name, url = sorted(year_files.items())[-1]

            rs_script = rs.make_script()
            if rs_script.get("a_file_name") != file_name:
                log(
                    f"Found new file {file_name} for rate script starting "
                    f"{date_format(year_start)}"
                )
                script = {"a_file_name": file_name}
                server_j = json.load(BytesIO(download(url)))
                script["ccl_gbp_per_kwh"] = Decimal(server_j["ccl_gbp_per_kwh"])
                rs.update(script)
                log(f"Updated CCL rate script for {date_format(year_start)}")

    log("Finished CCL rates")
    sess.commit()



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook
from openpyxl.utils import get_column_letter

from werkzeug.exceptions import BadRequest

from chellow.models import Session
from chellow.utils import parse_mpan_core, to_ct, to_utc

COLUMN_MAP = {}


def make_column_lookup(sheet):
    column_lookup = {}
    for cell in sheet[11]:
        title = str(cell.value).strip().lower()
        try:
            title = COLUMN_MAP[title]
        except KeyError:
            pass
        if title in column_lookup:
            title += " 2"
        column_lookup[title] = get_column_letter(cell.column)
    return column_lookup


def get_ct_date(sheet, col, row):
    cell = get_cell(sheet, col, row)
    val = cell.value
    if isinstance(val, Datetime):
        pass
    elif isinstance(val, str):
        if len(val) == 10:
            val = Datetime.strptime(val, "%d/%m/%Y")
        elif len(val) == 16:
            val = Datetime.strptime(val, "%d/%m/%Y %H:%M")
        else:
            raise BadRequest(
                f"Problem parsing string {val} as a timestamp at {cell.coordinate}."
            )
    else:
        raise BadRequest(f"Problem reading {val} as a timestamp at {cell.coordinate}.")
    return to_ct(val)


def get_start_date(sheet, col, row):
    dt = get_ct_date(sheet, col, row)
    return None if dt is None else to_utc(dt)


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_dec(sheet, col, row):
    cell = get_cell(sheet, col, row)
    if cell.value is None:
        return None
    else:
        try:
            return Decimal(str(cell.value))
        except InvalidOperation as e:
            raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_gbp(sheet, col, row):
    gbp = get_dec(sheet, col, row)
    if gbp is None:
        return gbp
    else:
        return Decimal("0.00") + round(gbp, 2)


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


def _process_row(sess, sheet, row, issue_date, cl):
    mpan_core = parse_mpan_core(str(get_int(sheet, "B", row)))
    start_date_ct = get_ct_date(sheet, cl["period from"], row)
    start_date = to_utc(start_date_ct)
    finish_date_ct = get_ct_date(sheet, cl["period to"], row) + relativedelta(
        hours=23, minutes=30
    )
    finish_date = to_utc(finish_date_ct)

    elements = []
    days = (finish_date_ct - start_date_ct).days + 1

    rate = get_dec(sheet, cl["dc pa"], row)
    net = get_gbp(sheet, cl["charge"], row)
    settlement_status_str = get_str(sheet, cl["type"], row)
    settlement_status = (
        "settlement" if settlement_status_str == "Settled" else "non-settlement"
    )
    comm = get_str(sheet, cl["comms"], row)
    elements.append(
        {
            "name": "mpan",
            "start_date": start_date,
            "finish_date": finish_date,
            "net": net,
            "breakdown": {
                "settlement-status": {settlement_status},
                "rate": {rate},
                "days": days,
                "comm": {comm},
            },
        }
    )

    vat = get_gbp(sheet, cl["est. vat"], row)
    gross = get_gbp(sheet, cl["est. total"], row)

    breakdown = {
        "raw_lines": [],
        "vat": {20: {"vat": vat, "net": net}},
    }

    return {
        "bill_type_code": "N",
        "kwh": Decimal(0),
        "net": net,
        "vat": vat,
        "gross": gross,
        "reads": [],
        "breakdown": breakdown,
        "account": mpan_core,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "mpan_core": mpan_core,
        "reference": "_".join(
            (
                start_date_ct.strftime("%Y%m%d"),
                finish_date_ct.strftime("%Y%m%d"),
                to_ct(issue_date).strftime("%Y%m%d"),
                mpan_core,
            )
        ),
        "elements": elements,
    }


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[1]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        try:
            with Session() as sess:
                column_lookup = make_column_lookup(self.sheet)
                bills = []
                issue_date = get_start_date(self.sheet, "C", 6)

                for row in range(12, len(self.sheet["B"]) + 1):
                    val = get_cell(self.sheet, "B", row).value
                    if val is None or val == "":
                        continue

                    self._set_last_line(row, val)
                    bill = _process_row(
                        sess, self.sheet, row, issue_date, column_lookup
                    )
                    bills.append(bill)
                    sess.rollback()

        except BadRequest as e:
            raise BadRequest(f"Row number: {row} {e.description}")

        return bills



from flask import g
from flask_restx import Api, Resource, fields, inputs, reqparse

from sqlalchemy import null, or_, select

import chellow.reports.report_247
from chellow.models import (
    Channel,
    Era,
    HhDatum,
    ReportRun,
    ReportRunRow,
    Site,
    Source,
    Supply,
)
from chellow.utils import to_utc

api = Api(
    version="1.0",
    title="Chellow API",
    description="Access Chellow data",
    endpoint="/api/v1",
    doc="/api/v1",
)

ns = api.namespace("v1", path="/api/v1/")

generator_type_model = ns.model(
    "GeneratorType",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="gen"),
        "description": fields.String(example=""),
    },
)

gsp_group_model = ns.model(
    "GspGroup",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="gen"),
        "name": fields.String(example=""),
    },
)

dno_group_model = ns.model(
    "Dno",
    {
        "id": fields.Integer(example="884"),
        "dno_code": fields.String(example="22"),
    },
)

sites__model = ns.model(
    "Site",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="GGJ23"),
        "name": fields.String(example="Trowbridge Waterworks"),
    },
)

site__era__model = ns.model(
    "Era",
    {
        "id": fields.Integer(example="81"),
        "supply_id": fields.Integer(example="988"),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "finish_date": fields.DateTime(example="2019-06-10T11:17:00+00:00"),
        "imp_mpan_core": fields.String,
        "exp_mpan_core": fields.String,
    },
)

site__site_era__model = ns.model(
    "SiteEra",
    {
        "id": fields.Integer(example="74"),
        "era": fields.Nested(site__era__model),
        "is_physical": fields.Boolean(example="true"),
    },
)
site_hh_datum_model = ns.model(
    "SiteHhDatum",
    {
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "imp_grid": fields.Float,
        "exp_grid": fields.Float,
        "imp_gen": fields.Float,
        "exp_gen": fields.Float,
        "imp_3p": fields.Float,
        "exp_3p": fields.Float,
        "used": fields.Float,
        "displaced": fields.Float,
    },
)

site__model = ns.model(
    "Site",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="GGJ23"),
        "name": fields.String(example="Trowbridge Waterworks"),
        "site_eras": fields.List(fields.Nested(site__site_era__model)),
        "hh_data": fields.List(fields.Nested(site_hh_datum_model)),
    },
)

source_model = ns.model(
    "Source",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="GGJ23"),
        "name": fields.String(example="Trowbridge Waterworks"),
    },
)
hh_data_model = ns.model(
    "HhData",
    {
        "from": fields.Integer,
        "to": fields.String,
        "name": fields.String,
    },
)

eras__site__model = ns.model(
    "Site",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="GGJ23"),
        "name": fields.String(example="Trowbridge Waterworks"),
    },
)

eras__site_eras__model = ns.model(
    "SiteEra",
    {
        "id": fields.Integer(example="74"),
        "site": fields.Nested(eras__site__model),
        "is_physical": fields.Boolean(example="true"),
    },
)

eras__mop_contract__model = ns.model(
    "MOP Contract",
    {
        "id": fields.Integer(example="821"),
        "name": fields.String(example="Marmaduke's Metering"),
    },
)

eras__dc_contract__model = ns.model(
    "DC Contract",
    {
        "id": fields.Integer(example="821"),
        "name": fields.String(example="Daisy's Data Collection"),
    },
)

pc__model = ns.model(
    "Profile Class",
    {
        "id": fields.Integer(example="8"),
        "code": fields.String(example="00"),
        "description": fields.String(example="Half-hourly"),
    },
)

era__participant__model = ns.model(
    "Participant",
    {
        "id": fields.Integer(example="91"),
        "code": fields.String(example="CUST"),
        "name": fields.String(example="Ingram Enterprises"),
    },
)

era__dno__model = ns.model(
    "DNO",
    {
        "id": fields.Integer(example="8"),
        "dno_code": fields.String(example="33"),
    },
)


era__mtc__model = ns.model(
    "Meter Timeswitch Class",
    {
        "id": fields.Integer(example="89"),
        "code": fields.String(example="800"),
        "description": fields.String(example="Unmetered"),
    },
)

cop__model = ns.model(
    "Code Of Practice",
    {
        "id": fields.Integer(example="89"),
        "code": fields.String(example="6c"),
        "description": fields.String(example="NHH"),
    },
)

eras__model = ns.model(
    "Eras",
    {
        "id": fields.Integer(example="712"),
        "supply_id": fields.Integer(example="349"),
        "site_eras": fields.List(fields.Nested(eras__site_eras__model)),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "finish_date": fields.DateTime(example="2019-06-10T11:17:00+00:00"),
    },
)


era__clock_interval__model = ns.model(
    "Clock Interval",
    {
        "id": fields.Integer(example="920"),
        "day_of_week": fields.Integer(example="2"),
        "start_month": fields.Integer(example="10"),
        "start_day": fields.Integer(example="23"),
        "start_hour": fields.Integer(example="23"),
        "start_minute": fields.Integer(example="30"),
        "end_month": fields.Integer(example="12"),
        "end_day": fields.Integer(example="13"),
        "end_hour": fields.Integer(example="6"),
        "end_minute": fields.Integer(example="0"),
    },
)

era__tpr__model = ns.model(
    "Tariff Pattern Regime",
    {
        "id": fields.Integer(example="71"),
        "code": fields.String(example="00001"),
        "is_teleswitch": fields.Boolean(example="false"),
        "is_gmt": fields.Boolean(example="true"),
        "clock_intervals": fields.List(fields.Nested(era__clock_interval__model)),
    },
)


measurement_requirement__model = ns.model(
    "Measurement Requirement",
    {
        "id": fields.Integer(example="72"),
        "tpr": fields.Nested(era__tpr__model),
    },
)

ssc__model = ns.model(
    "Standard Settlement Configuration",
    {
        "id": fields.Integer(example="11"),
        "code": fields.String(example="0363"),
        "description": fields.String(example="Unrestricted"),
        "is_import": fields.Boolean(example="true"),
        "valid_from": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "valid_to": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "measurement_requirements": fields.List(
            fields.Nested(measurement_requirement__model)
        ),
    },
)

energisation_status__model = ns.model(
    "Energisation Status",
    {
        "id": fields.Integer(example="25"),
        "code": fields.String(example="E"),
        "description": fields.String(example="Energised"),
    },
)

era__llfc__model = ns.model(
    "Line Loss Factor Class",
    {
        "id": fields.Integer(example="54"),
        "code": fields.String(example="010"),
        "description": fields.String(example="HV Sub"),
    },
)

era__supplier_contract__model = ns.model(
    "Supplier Contract",
    {
        "id": fields.Integer(example="871"),
        "name": fields.String(example="High Charges"),
    },
)


era__mop_contract__model = ns.model(
    "MOP Contract",
    {
        "id": fields.Integer(example="821"),
        "name": fields.String(example="Marmaduke's Metering"),
    },
)

hh_datum_model = ns.model(
    "HH Datum",
    {
        "id": fields.Integer(example="871"),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "value": fields.String(example="0.01"),
        "last_modified": fields.DateTime(example="2019-05-18T15:00:00+00:00"),
    },
)


def hh_datum_to_j(hh_datum):
    return {
        "id": hh_datum.id,
        "start_date": hh_datum.start_date,
        "value": hh_datum.value,
        "last_modified": hh_datum.last_modified,
    }


channel_model = ns.model(
    "Channel",
    {
        "id": fields.Integer(example="871"),
        "imp_related": fields.Boolean(example="true"),
        "channel_type": fields.String(example="ACTIVE"),
        "hh_data": fields.List(fields.Nested(hh_datum_model)),
    },
)


def channel_to_j(channel):
    return {
        "id": channel.id,
        "era_id": channel.era_id,
        "imp_related": channel.imp_related,
        "channel_type": channel.channel_type,
    }


era__model = ns.model(
    "Era",
    {
        "id": fields.Integer(example="712"),
        "supply_id": fields.Integer(example="349"),
        "site_eras": fields.List(fields.Nested(eras__site_eras__model)),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "finish_date": fields.DateTime(example="2019-06-10T11:17:00+00:00"),
        "mop_contract": fields.Nested(era__mop_contract__model),
        "mop_account": fields.String(example="xc771"),
        "dc_contract": fields.Nested(eras__dc_contract__model),
        "dc_account": fields.String(example="703n"),
        "msn": fields.String(example="mn88hg2"),
        "pc": fields.Nested(pc__model),
        "mtc": fields.Nested(era__mtc__model),
        "cop": fields.Nested(cop__model),
        "ssc": fields.Nested(ssc__model),
        "energisation_status": fields.Nested(energisation_status__model),
        "properties": fields.String,
        "imp_mpan_core": fields.String(example="22 5478 7392 943"),
        "imp_llfc": fields.Nested(era__llfc__model),
        "imp_supplier_contract": fields.Nested(era__supplier_contract__model),
        "imp_supplier_account": fields.String(example="x66g"),
        "imp_sc": fields.Integer(example="700"),
        "exp_mpan_core": fields.String(example="22 8672 2951 081"),
        "exp_llfc": fields.Nested(era__llfc__model),
        "exp_supplier_contract": fields.Nested(era__supplier_contract__model),
        "exp_supplier_account": fields.String(example="h8892"),
        "exp_sc": fields.Integer(example="900"),
        "channels": fields.List(fields.Nested(channel_model)),
    },
)


channel_model = ns.model(
    "Channel",
    {
        "id": fields.Integer(example="871"),
        "imp_related": fields.Boolean(example="true"),
        "channel_type": fields.String(example="ACTIVE"),
        "hh_data": fields.List(fields.Nested(hh_datum_model)),
    },
)


def report_run_row_to_j(report_run_row):
    return {
        "id": report_run_row.id,
        "report_run_id": report_run_row.report_run_id,
        "tab": report_run_row.tab,
        "data": report_run_row.data,
    }


report_run_row_model = ns.model(
    "ReportRunRow",
    {
        "id": fields.Integer(example="712"),
        "report_run_id": fields.Integer(example="782"),
        "tab": fields.String(example="sites"),
        "data": fields.Raw(),
    },
)


def report_run_to_j(report_run):
    return {
        "id": report_run.id,
        "date_created": report_run.date_created,
        "creator": report_run.creator,
        "name": report_run.name,
        "title": report_run.title,
        "state": report_run.state,
        "data": report_run.data,
    }


report_run_model = ns.model(
    "ReportRun",
    {
        "id": fields.Integer(example="712"),
        "date_created": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "creator": fields.String(example="H. G. Wells"),
        "name": fields.String(example="bill_check"),
        "title": fields.String(example="batch_55-00"),
        "state": fields.Raw(),
        "data": fields.Raw(),
        "rows": fields.List(fields.Nested(report_run_row_model)),
    },
)


def site_to_j(site):
    return {
        "id": site.id,
        "code": site.code,
        "name": site.name,
    }


@ns.route("/sites")
class SitesResource(Resource):
    @api.marshal_list_with(sites__model)
    def get(self):
        return g.sess.query(Site).order_by(Site.code).all()


site_parser = ns.parser()
site_parser.add_argument(
    "start_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Start of HH data",
    default="2025-12-01T00:00Z",
)
site_parser.add_argument(
    "finish_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Finish of HH data",
    default="2025-12-31T23:30Z",
)


@ns.route("/sites/<int:site_id>")
@ns.param("site_id", "The site identifier")
class SiteResource(Resource):
    @api.marshal_with(site__model)
    @ns.expect(site_parser)
    def get(self, site_id):
        args = site_parser.parse_args()
        start_date = to_utc(args["start_date"])
        finish_date = to_utc(args["finish_date"])
        site = Site.get_by_id(g.sess, site_id)
        site_j = site_to_j(site)
        site_j["hh_data"] = site.hh_data(
            g.sess, start_date, finish_date, exclude_virtual=True
        )
        return site_j


@ns.route("/eras/<int:era_id>")
class EraResource(Resource):
    @api.marshal_with(era__model)
    def get(self, era_id):
        return Era.get_by_id(g.sess, era_id)


channel_get_parser = reqparse.RequestParser()
channel_get_parser.add_argument(
    "from",
    type=inputs.datetime_from_iso8601,
    help="Beginning of date range for half-hourly data.",
)
channel_get_parser.add_argument(
    "to",
    type=inputs.datetime_from_iso8601,
    help="End of date range for half-hourly data.",
)


channel_parser = ns.parser()
channel_parser.add_argument(
    "start_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Start of HH data",
    default="2025-12-01T00:00Z",
)
channel_parser.add_argument(
    "finish_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Finish of HH data",
    default="2025-12-31T23:30Z",
)


@ns.route("/channel/<int:channel_id>")
class ChannelResource(Resource):
    @api.marshal_with(channel_model)
    @ns.expect(channel_parser)
    def get(self, channel_id):
        args = channel_parser.parse_args()
        start_date = to_utc(args["start_date"])
        finish_date = to_utc(args["finish_date"])
        channel = Channel.get_by_id(g.sess, channel_id)
        channel_j = channel_to_j(channel)
        hh_data_j = []
        for hh_datum in g.sess.scalars(
            select(HhDatum).where(
                HhDatum.channel == channel,
                HhDatum.start_date >= start_date,
                HhDatum.start_date <= finish_date,
            )
        ):
            hh_data_j.append(hh_datum_to_j(hh_datum))
        channel_j["hh_data"] = hh_data_j
        return channel_j


supply_model = ns.model(
    "Supply",
    {
        "id": fields.Integer(example="712"),
        "name": fields.String(example="hgeoiu"),
        "source": fields.Nested(source_model),
        "generator_type": fields.Nested(generator_type_model, required=False),
        "gsp_group": fields.Nested(gsp_group_model),
        "dno": fields.Nested(dno_group_model),
        "eras": fields.List(fields.Nested(era__model)),
    },
)

supplies_parser = ns.parser()
supplies_parser.add_argument(
    "start_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Start of eras, channels and HH data",
    default="2025-12-01T00:00Z",
)
supplies_parser.add_argument(
    "finish_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Finish of eras, channales and HH data",
    default="2025-12-31T23:30Z",
)
supplies_parser.add_argument("source_code", type=str, default="gen")


@ns.route("/supplies")
class SuppliesResource(Resource):
    @api.marshal_list_with(supply_model)
    @ns.expect(supplies_parser)
    def get(self):
        args = supplies_parser.parse_args()
        start_date = to_utc(args["start_date"])
        finish_date = to_utc(args["finish_date"])

        supplies_j = []
        supplies_q = (
            select(Supply)
            .join(Era)
            .where(
                Era.start_date <= finish_date,
                or_(Era.finish_date == null(), Era.finish_date >= start_date),
            )
            .distinct()
        )
        if "source_code" in args:
            supplies_q = supplies_q.join(Source).where(
                Source.code == args["source_code"]
            )
        for supply in g.sess.scalars(supplies_q):
            eras_j = []
            supply_j = {
                "id": supply.id,
                "name": supply.name,
                "source": supply.source,
                "generator_type": supply.generator_type,
                "gsp_group": supply.gsp_group,
                "dno": supply.dno,
                "eras": eras_j,
            }
            supplies_j.append(supply_j)
            for era in g.sess.scalars(
                select(Era)
                .where(
                    Era.supply == supply,
                    Era.start_date <= finish_date,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                )
                .order_by(Era.start_date)
            ):
                channels_j = []
                era_j = {
                    "id": era.id,
                    "site_eras": era.site_eras,
                    "start_date": era.start_date,
                    "finish_date": era.finish_date,
                    "mop_contract": era.mop_contract,
                    "dc_contract": era.dc_contract,
                    "msn": era.msn,
                    "pc": era.pc,
                    "mtc_participant": era.mtc_participant,
                    "cop": era.cop,
                    "ssc": era.ssc,
                    "energisation_status": era.energisation_status,
                    "imp_mpan_core": era.imp_mpan_core,
                    "imp_llfc": era.imp_llfc,
                    "imp_supplier_contract": era.imp_supplier_contract,
                    "imp_supplier_account": era.imp_supplier_account,
                    "imp_sc": era.imp_sc,
                    "exp_mpan_core": era.exp_mpan_core,
                    "exp_llfc": era.exp_llfc,
                    "exp_supplier_contract": era.exp_supplier_contract,
                    "exp_supplier_account": era.exp_supplier_account,
                    "exp_sc": era.exp_sc,
                    "channels": channels_j,
                }
                eras_j.append(era_j)
                for channel in era.channels:

                    channel_j = {
                        "id": channel.id,
                        "era_id": channel.era_id,
                        "imp_related": channel.imp_related,
                        "channel_type": channel.channel_type,
                    }
                    channels_j.append(channel_j)
        return supplies_j


supply_parser = ns.parser()
supply_parser.add_argument(
    "start_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Start of HH data",
    default="2025-12-01T00:00Z",
)
supply_parser.add_argument(
    "finish_date",
    type=inputs.datetime_from_iso8601,
    required=True,
    help="Finish of HH data",
    default="2025-12-31T23:30Z",
)


@ns.route("/supplies/<int:supply_id>")
class SupplyResource(Resource):
    @api.marshal_with(supply_model)
    @ns.expect(supply_parser)
    def get(self, supply_id):
        args = supply_parser.parse_args()
        start_date = to_utc(args["start_date"])
        finish_date = to_utc(args["finish_date"])

        supply = Supply.get_by_id(g.sess, supply_id)
        eras_j = []
        supply_j = {
            "id": supply.id,
            "name": supply.name,
            "source": supply.source,
            "generator_type": supply.generator_type,
            "gsp_group": supply.gsp_group,
            "dno": supply.dno,
            "eras": eras_j,
        }
        for era in g.sess.scalars(
            select(Era)
            .where(
                Era.supply == supply,
                Era.start_date <= finish_date,
                or_(Era.finish_date == null(), Era.finish_date >= start_date),
            )
            .order_by(Era.start_date)
        ):
            channels_j = []
            era_j = {
                "id": era.id,
                "site_eras": era.site_eras,
                "start_date": era.start_date,
                "finish_date": era.finish_date,
                "mop_contract": era.mop_contract,
                "dc_contract": era.dc_contract,
                "msn": era.msn,
                "pc": era.pc,
                "mtc_participant": era.mtc_participant,
                "cop": era.cop,
                "ssc": era.ssc,
                "energisation_status": era.energisation_status,
                "imp_mpan_core": era.imp_mpan_core,
                "imp_llfc": era.imp_llfc,
                "imp_supplier_contract": era.imp_supplier_contract,
                "imp_supplier_account": era.imp_supplier_account,
                "imp_sc": era.imp_sc,
                "exp_mpan_core": era.exp_mpan_core,
                "exp_llfc": era.exp_llfc,
                "exp_supplier_contract": era.exp_supplier_contract,
                "exp_supplier_account": era.exp_supplier_account,
                "exp_sc": era.exp_sc,
                "channels": channels_j,
            }
            eras_j.append(era_j)
            for channel in era.channels:

                channel_j = {
                    "id": channel.id,
                    "era_id": channel.era_id,
                    "imp_related": channel.imp_related,
                    "channel_type": channel.channel_type,
                }
                hh_data_j = []
                for hh_datum in g.sess.scalars(
                    select(HhDatum)
                    .where(
                        HhDatum.channel == channel,
                        HhDatum.start_date >= start_date,
                        HhDatum.start_date <= finish_date,
                    )
                    .order_by(HhDatum.start_date)
                ):
                    hh_datum_j = {
                        "id": hh_datum.id,
                        "start_date": hh_datum.start_date,
                        "value": hh_datum.value,
                        "last_modified": hh_datum.last_modified,
                    }
                    hh_data_j.append(hh_datum_j)
                channel_j["hh_data"] = hh_data_j

                channels_j.append(channel_j)
        return supply_j


@ns.route("/eras")
class ErasResource(Resource):
    @api.marshal_list_with(eras__model)
    def get(self):
        return (
            g.sess.query(Era).filter(Era.finish_date == null()).order_by(Era.id).all()
        )


report_model = ns.model(
    "Report",
    {
        "report_run_id": fields.Integer(example="712"),
    },
)

report_monthly_duration_parser = ns.parser()
report_monthly_duration_parser.add_argument(
    "finish_year",
    type=int,
    help="Finish year of period",
    default="2025",
)
report_monthly_duration_parser.add_argument(
    "finish_month",
    type=int,
    help="Finish month of period",
    default="12",
)
report_monthly_duration_parser.add_argument(
    "months",
    type=int,
    help="Duration of period",
    default="1",
)
report_monthly_duration_parser.add_argument(
    "site_code",
    action="append",
    type=str,
    help="Runs for these site codes. If no site codes specified, "
    "then runs for all sites",
)


@ns.route("/reports/monthly_duration")
class ReportMonthlyDurationResource(Resource):
    @api.marshal_with(report_model)
    @ns.expect(report_monthly_duration_parser)
    def get(self):
        args = report_monthly_duration_parser.parse_args()
        return chellow.reports.report_247.do_get_j(args)


report_runs_parser = ns.parser()


@ns.route("/report_runs")
class ReportRunsResource(Resource):
    @api.marshal_list_with(report_run_model)
    @ns.expect(report_runs_parser)
    def get(self):
        report_runs_parser.parse_args()

        report_runs_j = []
        for report_run in g.sess.scalars(select(ReportRun)):
            report_runs_j.append(report_run_to_j(report_run))
        return report_runs_j


report_run_parser = ns.parser()
report_run_parser.add_argument(
    "include_rows",
    type=inputs.boolean,
    help="Show report rows in the response",
    default="false",
)


@ns.route("/report_runs/<int:report_run_id>")
class ReportRunResource(Resource):
    @api.marshal_list_with(report_run_model)
    @ns.expect(report_run_parser)
    def get(self, report_run_id):
        report_run_parser.parse_args()
        report_run_j = report_run_to_j(ReportRun.get_by_id(g.sess, report_run_id))
        rows = []
        for row in g.sess.scalars(
            select(ReportRunRow).where(ReportRunRow.report_run_id == report_run_id)
        ):
            rows.append(report_run_row_to_j(row))
        report_run_j["rows"] = rows
        return report_run_j



from decimal import Decimal
from io import BytesIO
from zipfile import ZipFile

from openpyxl import Workbook

import pytest

from chellow.e.dno_rate_parser import (
    find_rates,
    str_to_hr,
    tab_ehv,
    tab_llfs,
    tab_lv_hv,
    to_llfcs,
    to_pcs,
)


@pytest.mark.parametrize(
    "llfc_str,llfc_list",
    [
        ["H00-H01", ["H00", "H01"]],
        ["310 - 311 inclusive", ["310", "311"]],
        [
            """310
311
""",
            ["310", "311"],
        ],
    ],
)
def test_to_llfcs(mocker, llfc_str, llfc_list):
    actual = to_llfcs(llfc_str)
    assert actual == llfc_list


@pytest.mark.parametrize(
    "pc_str,pc_list", [["3 to 8 or 0", ["03", "04", "05", "06", "07", "08", "00"]]]
)
def test_to_pcs(mocker, pc_str, pc_list):
    mocker.patch("chellow.e.dno_rate_parser.get_value", return_value=pc_str)

    row = mocker.Mock()
    idx = mocker.Mock()
    actual = to_pcs(row, idx)
    assert actual == pc_list


@pytest.mark.parametrize("hr_str,hr", [["1600", Decimal("16")]])
def test_str_to_hr(mocker, hr_str, hr):
    actual = str_to_hr(hr_str)
    assert actual == hr


def test_find_rates():
    file_name = "a.zip"
    file_like = BytesIO()
    with ZipFile(file_like, "w") as zf:
        zf.writestr("b.txt", "")
    file_like.seek(0)
    find_rates(file_name, file_like)


def test_tab_llfs():
    wb = Workbook()
    wb.create_sheet("Annex 5 LLFs")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 23)
    sheet.insert_cols(0, 9)
    sheet["A14"].value = "Metered voltage"
    sheet["A15"].value = "Low-voltage network"
    sheet["H15"].value = "100,ABB"

    vls = tab_llfs(sheet)
    assert vls == [
        {"code": "100", "voltage_level": "LV", "is_substation": False},
        {"code": "ABB", "voltage_level": "LV", "is_substation": False},
    ]


def test_tab_llfs_blank():
    wb = Workbook()
    wb.create_sheet("Annex 5 LLFs")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 23)
    sheet.insert_cols(0, 9)
    sheet["A14"].value = "Metered voltage"
    sheet["A15"].value = "Low-voltage network"
    sheet["H15"].value = None

    vls = tab_llfs(sheet)
    assert vls == []


def test_tab_lv_hv():
    wb = Workbook()
    wb.create_sheet("Annex 1 LV, HV and UMS charges")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 23)
    sheet.insert_cols(0, 12)
    sheet["A5"].value = "Time periods"
    sheet["A6"].value = "Saturday and Sunday\nAll year"
    sheet["B6"].value = ""
    sheet["C6"].value = "09:30 - 21:30"
    sheet["E6"].value = "00:00 - 09:30\n21:20 - 24:00"
    sheet["A7"].value = "Notes"

    sheet["G6"].value = (
        "Monday to Friday\n(Including Bank Holidays)\n Nov to Feb Inclusive "
        "(excluding 22nd Dec to 4th Jan inclusive)"
    )
    sheet["H6"].value = ""
    sheet["I6"].value = "17:00 to 19:00"
    sheet["J6"].value = "07:30 to 17:00\n19:00 to 21:30"
    sheet["G7"].value = "Notes"

    rates = {}
    tab_lv_hv(sheet, rates)
    assert rates == {
        "bands": [
            {
                "band": "amber",
                "finish": Decimal("21.5"),
                "start": Decimal("9.5"),
                "weekend": True,
            }
        ],
        "ums_bands": [
            {
                "band": "black",
                "finish-month": 12,
                "finish-day": 21,
                "finish-decimal-hour": Decimal("19"),
                "start-month": 11,
                "start-day": 1,
                "start-decimal-hour": Decimal("17"),
                "weekend": False,
            },
            {
                "band": "black",
                "finish-day": "last",
                "finish-decimal-hour": Decimal("19"),
                "finish-month": 2,
                "start-day": 4,
                "start-decimal-hour": Decimal("17"),
                "start-month": 1,
                "weekend": False,
            },
            {
                "band": "yellow",
                "finish-day": 21,
                "finish-decimal-hour": Decimal("17"),
                "finish-month": 12,
                "start-day": 1,
                "start-decimal-hour": Decimal("7.5"),
                "start-month": 11,
                "weekend": False,
            },
            {
                "band": "yellow",
                "finish-day": "last",
                "finish-decimal-hour": Decimal("17"),
                "finish-month": 2,
                "start-day": 4,
                "start-decimal-hour": Decimal("7.5"),
                "start-month": 1,
                "weekend": False,
            },
            {
                "band": "yellow",
                "finish-day": 21,
                "finish-decimal-hour": Decimal("21.5"),
                "finish-month": 12,
                "start-day": 1,
                "start-decimal-hour": Decimal("19"),
                "start-month": 11,
                "weekend": False,
            },
            {
                "band": "yellow",
                "finish-day": "last",
                "finish-decimal-hour": Decimal("21.5"),
                "finish-month": 2,
                "start-day": 4,
                "start-decimal-hour": Decimal("19"),
                "start-month": 1,
                "weekend": False,
            },
        ],
        "tariffs": {},
    }


def test_tab_ehv():
    wb = Workbook()
    wb.create_sheet("Annex 2 EHV charges")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 10)
    sheet.insert_cols(0, 15)
    sheet["A5"].value = "time periods"
    sheet["A6"].value = (
        "Monday to Friday (Including Bank Holidays) Nov to Feb Inclusive (excluding "
        "22nd Dec to 4th Jan inclusive)"
    )
    sheet["D6"].value = "17:00 to 19:00"
    sheet["A7"].value = "Notes"
    sheet["A10"].value = "Import Unique Identifier"
    sheet["B10"].value = "LLFC"
    sheet["E10"].value = "LLFC"
    sheet["H10"].value = "Residual Charging Band"
    sheet["I10"].value = "Import Super Red unit charge (p/kWh)"
    sheet["J10"].value = "Import fixed charge (p/day)"
    sheet["K10"].value = "Import capacity charge (p/kVA/day)"
    sheet["L10"].value = "Import exceeded capacity charge (p/kVA/day)"
    sheet["B11"].value = "198"
    sheet["H11"].value = "2"
    sheet["I11"].value = "4.124"
    sheet["J11"].value = "18767.69"
    sheet["K11"].value = "1.40"
    sheet["L11"].value = "1.40"

    rates = {}
    tab_ehv(sheet, rates)
    assert rates == {
        "tariffs": {
            "198": {
                "super-red-gbp-per-kwh": Decimal("0.04124"),
                "gbp-per-mpan-per-day": Decimal("187.67690"),
                "gbp-per-kva-per-day": Decimal("0.01400"),
                "excess-gbp-per-kva-per-day": Decimal("0.01400"),
                "description": "Designated EHV2",
            }
        },
        "super_red": [
            {
                "finish-day": 21,
                "finish-decimal-hour": Decimal("19"),
                "finish-month": 12,
                "start-day": 1,
                "start-decimal-hour": Decimal("17"),
                "start-month": 11,
                "weekend": False,
            },
            {
                "finish-day": "last",
                "finish-decimal-hour": Decimal("19"),
                "finish-month": 2,
                "start-day": 4,
                "start-decimal-hour": Decimal("17"),
                "start-month": 1,
                "weekend": False,
            },
        ],
    }


def test_tab_ehv_null_band():
    wb = Workbook()
    wb.create_sheet("Annex 2 EHV charges")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 10)
    sheet.insert_cols(0, 15)
    sheet["A10"].value = "Import Unique Identifier"
    sheet["B10"].value = "LLFC"
    sheet["E10"].value = "LLFC"
    sheet["H10"].value = "Residual Charging Band"
    sheet["I10"].value = "Import Super Red unit charge (p/kWh)"
    sheet["J10"].value = "Import fixed charge (p/day)"
    sheet["K10"].value = "Import capacity charge (p/kVA/day)"
    sheet["L10"].value = "Import exceeded capacity charge (p/kVA/day)"
    sheet["B11"].value = "198"
    sheet["H11"].value = None
    sheet["I11"].value = "4.124"
    sheet["J11"].value = "18767.69"
    sheet["K11"].value = "1.40"
    sheet["L11"].value = "1.40"

    rates = {}
    tab_ehv(sheet, rates)
    assert rates == {
        "tariffs": {
            "198": {
                "super-red-gbp-per-kwh": Decimal("0.04124"),
                "gbp-per-mpan-per-day": Decimal("187.67690"),
                "gbp-per-kva-per-day": Decimal("0.01400"),
                "excess-gbp-per-kva-per-day": Decimal("0.01400"),
                "description": "Designated EHV0",
            }
        },
        "super_red": [],
    }



import re
from time import sleep


def match(response, status_code, *patterns):
    response_str = str(response.headers) + "\n" + response.get_data(as_text=True)

    assert response.status_code == status_code, response_str

    for regex in patterns:
        assert re.search(
            regex, response_str, flags=re.MULTILINE + re.DOTALL
        ), response_str


def match_tables(expected_table, actual_table):
    len_expected_table = len(expected_table)
    len_actual_table = len(actual_table)
    assert len_expected_table == len_actual_table, (
        f"The length of the expected table {len_expected_table} does not match the "
        f"lngth of the actual table {len_actual_table}"
    )
    for i, (expected_row, actual_row) in enumerate(zip(expected_table, actual_table)):
        len_expected_row, len_actual_row = len(expected_row), len(actual_row)
        assert len_expected_row == len_actual_row, (
            f"For row {i} the length {len_expected_row} of the expected row does not "
            f"match the lngth of the actual row {len_actual_row}\n"
            f"{expected_row}\n"
            f"{actual_row}\n"
        )
        for j, (expected_val, actual_val) in enumerate(zip(expected_row, actual_row)):
            assert expected_val == actual_val, (
                f"On row {i} column {j}, {expected_val} != {actual_val}\n"
                f"{expected_row}\n"
                f"{actual_row}\n"
            )


def match_repeat(client, path, match, seconds=5):
    for second in range(seconds):
        response = client.get(path)
        response_str = response.get_data(as_text=True)
        if match in response_str:
            return response
        elif second == seconds - 1:
            assert match in response_str, response_str
        else:
            sleep(1)



import csv
from decimal import Decimal
from io import StringIO

from utils import match

from chellow.e.computer import contract_func
from chellow.models import (
    BillType,
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    ReportRun,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_111 import (
    _process_period,
    _process_supply,
    content,
    find_gaps,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_find_gaps():
    gaps = {
        to_utc(ct_datetime(2025, 1, 1)): "start",
        to_utc(ct_datetime(2025, 1, 1, 0, 30)): "start_finish",
        to_utc(ct_datetime(2025, 1, 1, 1, 0)): "finish",
    }
    assert list(find_gaps(gaps)) == [
        (
            to_utc(ct_datetime(2025, 1, 1, 0, 0)),
            to_utc(ct_datetime(2025, 1, 1, 0, 0)),
        ),
        (
            to_utc(ct_datetime(2025, 1, 1, 0, 30)),
            to_utc(ct_datetime(2025, 1, 1, 0, 30)),
        ),
    ]


# End to end tests


def test_ete_error_message_for_invalid_bill_id(client, sess):
    query_string = {
        "bill_id": "0",
    }
    response = client.get("/reports/111", query_string=query_string)

    match(response, 404)


# HTTP level tests


def test_http_supplier_batch_with_mpan_cores(mocker, client, sess):
    vf = utc_datetime(2000, 1, 1)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    participant = Participant.insert(sess, "hhak", "AK Industries")
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    batch = supplier_contract.insert_batch(sess, "005", "batch 5", vf)
    sess.commit()
    user = User.get_by_email_address(sess, "admin@example.com")
    MockThread = mocker.patch("chellow.reports.report_111.threading.Thread")

    query_string = {
        "batch_id": str(batch.id),
        "mpan_cores": "22 1065 3921 534",
    }
    response = client.get("/reports/111", query_string=query_string)

    match(response, 303)

    expected_args = (
        batch.id,
        None,
        None,
        None,
        None,
        user.id,
        ["22 1065 3921 534"],
        "_batch_005",
        1,
    )

    MockThread.assert_called_with(target=content, args=expected_args)


# Worker level tests


def test_process_period(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill(ds):
    rate = 0.1
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['nrg-kwh'] = hh['msp-kwh']
        bill_hh['nrg-rate'] = {rate}
        bill_hh['off-rate'] = {0.1}
        bill_hh['nrg-gbp'] = hh['msp-kwh'] * rate + 1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']
        bill_hh['problem'] = 'hello '

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a b", "", vf)
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "dd",
        "hjk",
        to_utc(ct_datetime(2009, 7, 10)),
        to_utc(ct_datetime(2009, 7, 10)),
        to_utc(ct_datetime(2009, 7, 10)),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        bill_type,
        {"vat": {Decimal("20"): {"net": Decimal("10.00"), "vat": Decimal("10.00")}}},
        supply,
    )
    bill.insert_element(
        sess,
        "nrg",
        to_utc(ct_datetime(2009, 7, 10)),
        to_utc(ct_datetime(2009, 7, 10)),
        Decimal("10.00"),
        {"rate": {Decimal("0.1")}},
    )
    sess.commit()

    report_context = {}
    bill_ids = {bill.id: bill}
    forecast_date = utc_datetime(2020, 7, 10)
    vbf = contract_func(report_context, supplier_contract, "virtual_bill")
    period_start = to_utc(ct_datetime(2009, 7, 10, 0, 0))
    period_finish = to_utc(ct_datetime(2009, 7, 10, 0, 0))

    data = _process_period(
        sess,
        report_context,
        supply,
        supplier_contract,
        bill_ids,
        forecast_date,
        vbf,
        period_start,
        period_finish,
    )
    expected = {
        "actual_bills": [
            {
                "batch_id": 1,
                "batch_reference": "a b",
                "breakdown": "{\n"
                '  "vat": {\n'
                "    20: {\n"
                '      "net": 10.00,\n'
                '      "vat": 10.00,\n'
                "    },\n"
                "  },\n"
                "}",
                "gross": Decimal("10.00"),
                "id": 1,
                "kwh": Decimal("10.00"),
                "net": Decimal("10.00"),
                "problem": "The Gross GBP (10.00) of the bill isn't equal "
                "to the Net GBP "
                "(10.00) + VAT GBP (10.00) of the bill.",
                "start_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                "finish_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                "vat": Decimal("10.00"),
            },
        ],
        "actual_net_gbp": 10.0,
        "difference_net_gbp": 9.0,
        "contract_id": 4,
        "contract_name": "Fusion Supplier 2000",
        "market_role_code": "X",
        "elements": {
            "nrg": {
                "parts": {
                    "gbp": {
                        "actual": Decimal("10.00"),
                        "difference": 9.0,
                        "virtual": 1.0,
                        "passed": "❌",
                    },
                    "kwh": {
                        "difference": None,
                        "virtual": 0,
                        "passed": "❔",
                    },
                    "rate": {
                        "actual": {
                            Decimal("0.1"),
                        },
                        "difference": 0.0,
                        "virtual": {
                            0.1,
                        },
                        "passed": "✅",
                    },
                },
                "actual_elements": [
                    {
                        "bill": {
                            "batch": {
                                "id": 1,
                                "reference": "a b",
                            },
                            "id": 1,
                        },
                        "breakdown": "{\n" '  "rate": [\n' "    0.1,\n" "  ],\n" "}",
                        "finish_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                        "id": 1,
                        "net": Decimal("10.00"),
                        "start_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                    },
                ],
            },
        },
        "exp_mpan_core": None,
        "imp_mpan_core": "22 7867 6232 781",
        "period_start": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
        "period_finish": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
        "site_id": 1,
        "site_code": "CI017",
        "site_name": "Water Works",
        "supply_id": 1,
        "virtual_net_gbp": 1.0,
        "problem": "Bills have problems. hello ",
    }

    assert data == expected


def test_process_period_no_era(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill(ds):
    rate = 0.1
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['nrg-kwh'] = hh['msp-kwh']
        bill_hh['nrg-rate'] = {rate}
        bill_hh['off-rate'] = {0.1}
        bill_hh['nrg-gbp'] = hh['msp-kwh'] * rate + 1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']
        bill_hh['problem'] = 'hello '

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        to_utc(ct_datetime(2001, 1, 1)),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a b", "", vf)
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "dd",
        "hjk",
        to_utc(ct_datetime(2009, 7, 10)),
        to_utc(ct_datetime(2009, 7, 10)),
        to_utc(ct_datetime(2009, 7, 10)),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        bill_type,
        {"vat": {Decimal("20"): {"net": Decimal("10.00"), "vat": Decimal("10.00")}}},
        supply,
    )
    bill.insert_element(
        sess,
        "nrg",
        to_utc(ct_datetime(2009, 7, 10)),
        to_utc(ct_datetime(2009, 7, 10)),
        Decimal("10.00"),
        {"rate": {Decimal("0.1")}},
    )
    sess.commit()

    report_context = {}
    bill_ids = {bill.id: bill}
    forecast_date = utc_datetime(2020, 7, 10)
    vbf = contract_func(report_context, supplier_contract, "virtual_bill")
    period_start = to_utc(ct_datetime(2009, 7, 10, 0, 0))
    period_finish = to_utc(ct_datetime(2009, 7, 10, 0, 0))

    data = _process_period(
        sess,
        report_context,
        supply,
        supplier_contract,
        bill_ids,
        forecast_date,
        vbf,
        period_start,
        period_finish,
    )
    expected = {
        "actual_bills": [
            {
                "batch_id": 1,
                "batch_reference": "a b",
                "breakdown": "{\n"
                '  "vat": {\n'
                "    20: {\n"
                '      "net": 10.00,\n'
                '      "vat": 10.00,\n'
                "    },\n"
                "  },\n"
                "}",
                "gross": Decimal("10.00"),
                "id": 1,
                "kwh": Decimal("10.00"),
                "net": Decimal("10.00"),
                "problem": "The Gross GBP (10.00) of the bill isn't equal "
                "to the Net GBP (10.00) + VAT GBP (10.00) of the bill.",
                "start_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                "finish_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                "vat": Decimal("10.00"),
            },
        ],
        "actual_net_gbp": 10.0,
        "difference_net_gbp": 10.0,
        "contract_id": 4,
        "contract_name": "Fusion Supplier 2000",
        "market_role_code": "X",
        "elements": {
            "nrg": {
                "parts": {
                    "gbp": {
                        "actual": Decimal("10.00"),
                        "difference": 10.0,
                        "passed": "❌",
                    },
                    "rate": {
                        "actual": {
                            Decimal("0.1"),
                        },
                        "difference": None,
                        "passed": "❔",
                    },
                },
                "actual_elements": [
                    {
                        "bill": {
                            "batch": {
                                "id": 1,
                                "reference": "a b",
                            },
                            "id": 1,
                        },
                        "breakdown": "{\n" '  "rate": [\n' "    0.1,\n" "  ],\n" "}",
                        "finish_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                        "id": 1,
                        "net": Decimal("10.00"),
                        "start_date": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
                    },
                ],
            },
        },
        "exp_mpan_core": None,
        "imp_mpan_core": "22 7867 6232 781",
        "period_start": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
        "period_finish": to_utc(ct_datetime(2009, 7, 10, 0, 0)),
        "site_id": 1,
        "site_code": "CI017",
        "site_name": "Water Works",
        "supply_id": 1,
        "virtual_net_gbp": 0,
        "problem": "Bills have problems. No eras for this period of the supply. ",
    }

    assert data == expected


def test_process_supply(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill(ds):
    rate = 0.1
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['nrg-kwh'] = hh['msp-kwh']
        bill_hh['nrg-rate'] = {rate}
        bill_hh['off-rate'] = {0.1}
        bill_hh['nrg-gbp'] = hh['msp-kwh'] * rate + 1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']
        bill_hh['problem'] = 'hello '

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a b", "", vf)
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "dd",
        "hjk",
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        bill_type,
        {},
        supply,
    )
    bill.insert_element(
        sess,
        "nrg",
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        Decimal("10.00"),
        {"rate": {Decimal("0.1")}},
    )
    sess.commit()

    report_context = {}
    supply_id = supply.id
    bill_ids = {bill.id}
    forecast_date = utc_datetime(2020, 7, 10)
    vbf = contract_func(report_context, supplier_contract, "virtual_bill")

    vals = list(
        _process_supply(
            sess,
            report_context,
            supply_id,
            bill_ids,
            forecast_date,
            supplier_contract,
            vbf,
        )
    )
    expected = [
        {
            "actual_bills": [
                {
                    "batch_id": 1,
                    "batch_reference": "a b",
                    "breakdown": "{}",
                    "gross": Decimal("10.00"),
                    "id": 1,
                    "kwh": Decimal("10.00"),
                    "net": Decimal("10.00"),
                    "problem": "The Gross GBP (10.00) of the bill isn't equal "
                    "to the Net GBP "
                    "(10.00) + VAT GBP (10.00) of the bill.The total 'net' 0.00 in "
                    "the VAT breakdown doesn't match the 'net' 10.00 of the "
                    "bill.The total VAT 0.00 in the VAT breakdown doesn't match the "
                    "VAT 10.00 of the bill.",
                    "vat": Decimal("10.00"),
                    "start_date": utc_datetime(2009, 7, 10, 0, 0),
                    "finish_date": utc_datetime(2009, 7, 10, 0, 0),
                },
            ],
            "actual_net_gbp": 10.0,
            "difference_net_gbp": 9.0,
            "contract_id": 4,
            "contract_name": "Fusion Supplier 2000",
            "market_role_code": "X",
            "elements": {
                "nrg": {
                    "parts": {
                        "gbp": {
                            "actual": Decimal("10.00"),
                            "difference": 9.0,
                            "virtual": 1.0,
                            "passed": "❌",
                        },
                        "kwh": {
                            "difference": None,
                            "virtual": 0,
                            "passed": "❔",
                        },
                        "rate": {
                            "actual": {
                                Decimal("0.1"),
                            },
                            "difference": 0.0,
                            "virtual": {
                                0.1,
                            },
                            "passed": "✅",
                        },
                    },
                    "actual_elements": [
                        {
                            "bill": {
                                "batch": {
                                    "id": 1,
                                    "reference": "a b",
                                },
                                "id": 1,
                            },
                            "breakdown": "{\n"
                            '  "rate": [\n'
                            "    0.1,\n"
                            "  ],\n"
                            "}",
                            "finish_date": utc_datetime(2009, 7, 10, 0, 0),
                            "id": 1,
                            "net": Decimal("10.00"),
                            "start_date": utc_datetime(2009, 7, 10, 0, 0),
                        },
                    ],
                },
            },
            "exp_mpan_core": None,
            "imp_mpan_core": "22 7867 6232 781",
            "period_start": utc_datetime(2009, 7, 10, 0, 0),
            "period_finish": utc_datetime(2009, 7, 10, 0, 0),
            "site_id": 1,
            "site_code": "CI017",
            "site_name": "Water Works",
            "supply_id": 1,
            "virtual_net_gbp": 1.0,
            "problem": "Bills have problems. hello ",
        },
    ]

    assert vals == expected


def test_content(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        vf,
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill(ds):
    rate = 0.1
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['rate'] = {rate}
        bill_hh['off-rate'] = {0.1}
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * rate
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a b", "", vf)
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "dd",
        "hjk",
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        bill_type,
        {"rate": [Decimal("0.1")]},
        supply,
    )
    user_role = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "admin", user_role, None)
    user_id = user.id
    report_run = ReportRun.insert(sess, "", user, "", {})
    report_run_id = report_run.id
    batch_id = batch.id
    bill_id = bill.id
    supplier_contract_id = supplier_contract.id
    sess.commit()

    virtual_bill_titles = [
        "ccl-kwh",
        "ccl-rate",
        "ccl-gbp",
        "net-gbp",
        "vat-gbp",
        "gross-gbp",
        "sum-msp-kwh",
        "rate",
        "sum-msp-gbp",
        "problem",
    ]
    titles = [
        "batch",
        "bill-reference",
        "bill-type",
        "bill-kwh",
        "bill-net-gbp",
        "bill-vat-gbp",
        "bill-start-date",
        "bill-finish-date",
        "imp-mpan-core",
        "exp-mpan-core",
        "site-code",
        "site-name",
        "covered-from",
        "covered-to",
        "covered-bills",
        "metered-kwh",
    ]
    for t in virtual_bill_titles:
        titles.append("covered-" + t)
        titles.append("virtual-" + t)
        if t.endswith("-gbp"):
            titles.append("difference-" + t)

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_111.open_file", return_value=mock_file)

    content(
        batch_id,
        bill_id,
        supplier_contract_id,
        vf,
        vf,
        user_id,
        [],
        "",
        report_run_id,
    )


def test_content_with_mpans(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        vf,
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill_titles():
    return []

def virtual_bill(ds):
    rate = 0.1
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['rate'] = {rate}
        bill_hh['off-rate'] = {0.1}
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * rate
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a b", "", vf)
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "dd",
        "hjk",
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        bill_type,
        {"rate": [Decimal("0.1")]},
        supply,
    )
    user_role = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "admin", user_role, None)
    user_id = user.id
    report_run = ReportRun.insert(sess, "", user, "", {})
    report_run_id = report_run.id
    batch_id = batch.id
    bill_id = bill.id
    supplier_contract_id = supplier_contract.id
    sess.commit()

    virtual_bill_titles = [
        "ccl-kwh",
        "ccl-rate",
        "ccl-gbp",
        "net-gbp",
        "vat-gbp",
        "gross-gbp",
        "sum-msp-kwh",
        "rate",
        "sum-msp-gbp",
        "problem",
    ]
    titles = [
        "batch",
        "bill-reference",
        "bill-type",
        "bill-kwh",
        "bill-net-gbp",
        "bill-vat-gbp",
        "bill-start-date",
        "bill-finish-date",
        "imp-mpan-core",
        "exp-mpan-core",
        "site-code",
        "site-name",
        "covered-from",
        "covered-to",
        "covered-bills",
        "metered-kwh",
    ]
    for t in virtual_bill_titles:
        titles.append("covered-" + t)
        titles.append("virtual-" + t)
        if t.endswith("-gbp"):
            titles.append("difference-" + t)

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_111.open_file", return_value=mock_file)

    content(
        batch_id,
        bill_id,
        supplier_contract_id,
        vf,
        vf,
        user_id,
        [
            "22 7867 6232 781",
        ],
        "",
        report_run_id,
    )
    mock_file.seek(0)
    actual_rows = [row for row in csv.reader(mock_file)]
    assert actual_rows == [
        [
            "imp_mpan_core",
            "exp_mpan_core",
            "site_code",
            "site_name",
            "period_start",
            "period_finish",
            "actual_net_gbp",
            "virtual_net_gbp",
            "difference_net_gbp",
        ],
        [
            "22 7867 6232 781",
            "",
            "CI017",
            "Water Works",
            "2009-07-10 01:00",
            "2009-07-10 01:00",
            "0",
            "0.0",
            "0.0",
        ],
    ]



from chellow.e.elexon import run_import
from chellow.models import MarketRole, Participant
from chellow.utils import ct_datetime, to_utc


def test_elexon_import(mocker, sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    sess.commit()
    mocker.patch("chellow.e.system_price.elexon_import", autospec=True)
    mocker.patch("chellow.e.tlms.elexon_import", autospec=True)
    mocker.patch("chellow.e.rcrc.elexon_import", autospec=True)
    mocker.patch("chellow.e.lafs.elexon_import", autospec=True)

    log = mocker.Mock()
    set_progress = mocker.Mock()
    scripting_key = mocker.Mock()

    run_import(sess, log, set_progress, scripting_key)



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from werkzeug.exceptions import BadRequest

from chellow.utils import ct_datetime, to_ct, to_utc


def parse_edi(edi_str):
    for line_number, raw_line in enumerate(edi_str.split("'"), start=1):
        line = raw_line.strip()
        if len(line) == 0:
            continue
        code = line[:3]

        els = [el.split(":") for el in line[4:].split("+")]

        segment_name = code + els[1][0].strip() if code == "CCD" else code

        try:
            elem_data = SEGMENTS[segment_name]
        except KeyError:
            raise BadRequest(
                f"At line number {line_number} the segment name {segment_name} isn't "
                f"recognized. {raw_line}"
            )
        elem_codes = [m["code"] for m in elem_data["elements"]]

        elements = dict(zip(elem_codes, els))

        yield line_number, f"{line}'", segment_name, elements


def to_decimal(components):
    try:
        result = Decimal(components[0])
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing decimal '{components[0]}' {e}")
    if len(components) > 1 and components[-1] == "R":
        result *= Decimal("-1")
    return result


def to_gbp(components):
    return Decimal("0.00") + round(to_decimal(components) / Decimal("100"), 2)


def to_ct_date(component):
    if len(component) == 0:
        return None
    return to_ct(Datetime.strptime(component, "%y%m%d"))


def to_date(component):
    dt = to_ct_date(component)
    if dt is None:
        return None
    return to_utc(dt)


def to_finish_date(component):
    d = to_ct_date(component)
    return to_utc(ct_datetime(d.year, d.month, d.day, 23, 30))


def to_int(component):
    return int(component)


SEGMENTS = {
    "ADJ": {
        "description": "ADDITIONAL ADJUSTMENTS",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                ],
            },
            {
                "code": "SEQB",
                "description": "Second Level Sequence Number",
                "components": [
                    ("Second Level Sequence Number", "X"),
                ],
            },
            {
                "code": "ADJF",
                "description": "Adjustment Factor",
                "components": [
                    ("Adjustment Factor Code", "X"),
                    ("Adjustment Factor Value", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
        ],
    },
    "BTL": {
        "description": "BILL TRAILER",
        "elements": [
            {
                "code": "PTOT",
                "description": "Total of Payment Details",
                "components": [
                    ("Total of Payment Details", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "UVLT",
                "description": "Total Charge for Premises Before VAT",
                "components": [
                    ("Total Charge Before VAT", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "UTVA",
                "description": "Bill Total VAT Amount Payable",
                "components": [
                    ("Bill Total VAT Amount Payable", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "BABF",
                "description": "Balance Brought Forward",
                "components": [
                    ("Balance Brought Forward", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "TBTL",
                "description": "Total Bill Amount Payable",
                "components": [
                    ("Total Bill Amount Payable", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
        ],
    },
    "DNA": {
        "description": "DATA NARRATIVE",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [("First Level Sequence Number", "X")],
            },
            {
                "code": "DNAC",
                "description": "Data Narrative Code",
                "components": [("Code Table Number", "X"), ("Code Value", "X")],
            },
            {
                "code": "RTEX",
                "description": "Registered Text",
                "components": [
                    ("First Line Registered Text Code", "X"),
                    ("Application Text", "X"),
                    ("Second Line Registered Text Code", "X"),
                    ("Application Text", "X"),
                    ("Third Line Registered Text Code", "X"),
                    ("Application Text", "X"),
                    ("Fourth Line Registered Text Code", "X"),
                    ("Application Text", "X"),
                ],
            },
            {
                "code": "GNAR",
                "description": "General Narrative",
                "components": [
                    ("General Narrative Line 1", "X"),
                    ("General Narrative Line 2", "X"),
                    ("General Narrative Line 3", "X"),
                    ("General Narrative Line 4", "X"),
                ],
            },
        ],
    },
    "END": {
        "description": "END OF TRANSMISSION",
        "elements": [
            {
                "code": "NMST",
                "description": "Number of Messages in Transmission",
                "components": [("Number of Messages in Transmission", "X")],
            },
        ],
    },
    "STX": {
        "description": "Start Of Transmission",
        "elements": [
            {
                "code": "STDS",
                "description": "Syntax Rules Identifier",
                "components": [("Identifier", "X"), ("Version", "X")],
            },
            {
                "code": "FROM",
                "description": "Identification of Transmission Sender",
                "components": [("Code", "X"), ("Name", "X")],
            },
            {
                "code": "UNTO",
                "description": "Identification of Transmission Recipient",
                "components": [("Code", "X"), ("Name", "X")],
            },
            {
                "code": "TRDT",
                "description": "Date and Time of Transmission",
                "components": [("Date", "date"), ("Time", "time")],
            },
            {
                "code": "SNRF",
                "description": "Sender's Transmission Reference",
                "components": [("Sender's Transmission Reference", "X")],
            },
            {
                "code": "RCRF",
                "description": "Recipient's Transmission Reference",
                "components": [("Recipient's Transmission Reference", "X")],
            },
            {
                "code": "APRF",
                "description": "Application Reference",
                "components": [("Application Reference", "X")],
            },
            {
                "code": "PRCD",
                "description": "Transmission Priority Code",
                "components": [("Transmission Priority Code", "X")],
            },
        ],
    },
    "MHD": {
        "description": "MESSAGE HEADER",
        "elements": [
            {
                "code": "MSRF",
                "description": "Message Reference",
                "components": [("Message Reference", "X")],
            },
            {
                "code": "TYPE",
                "description": "Type of Message",
                "components": [("Type", "X"), ("Version Number", "X")],
            },
        ],
    },
    "TTL": {
        "description": "UTILITY BILL FILE TOTALS",
        "elements": [
            {
                "code": "FASU",
                "description": "Bill File Total Amount before VAT",
                "components": [
                    ("Bill File Total Amount before VAT", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "UVAT",
                "description": "Bill File Total VAT Amount",
                "components": [
                    ("Bill File Total VAT Amount", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "FTOP",
                "description": "File Total of Payment Details",
                "components": [
                    ("File Total of Payment Details", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "FBAB",
                "description": "File Total Balance Brought Forward",
                "components": [
                    ("File Total Balance Brought Forward", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "FPSU",
                "description": "Bill File Total Payable including VAT",
                "components": [
                    ("Bill File Total Payable including VAT", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "FTNI",
                "description": "File Total Number of Invoices",
                "components": [
                    ("File Total Number of Invoices", "X"),
                ],
            },
        ],
    },
    "TYP": {
        "description": "TRANSACTION TYPE DETAILS",
        "elements": [
            {
                "code": "TCDE",
                "description": "Transaction Code",
                "components": [("Transaction Code", "X")],
            },
            {
                "code": "TTYP",
                "description": "Transaction Type",
                "components": [("Transaction Type", "X")],
            },
        ],
    },
    "SDT": {
        "description": "SUPPLIER DETAILS",
        "elements": [
            {
                "code": "SIDN",
                "description": "Supplier's Identity",
                "components": [
                    ("Supplier's EAN Location Number", "X"),
                    ("Supplier's Identity Allocated by Customer", "X"),
                ],
            },
            {
                "code": "SNAM",
                "description": "Supplier's Name",
                "components": [("Supplier's Name", "X")],
            },
            {
                "code": "SADD",
                "description": "Supplier's Address",
                "components": [
                    ("Supplier's Address Line 1", "X"),
                    ("Supplier's Address Line 2", "X"),
                    ("Supplier's Address Line 3", "X"),
                    ("Supplier's Address Line 4", "X"),
                    ("Supplier's Post Code", "X"),
                ],
            },
            {
                "code": "VATN",
                "description": "Suplier's VAT Registration Number",
                "components": [
                    ("Numeric VAT Registration Number", "X"),
                    ("Alphanumeric VAT Registration Number", "X"),
                ],
            },
        ],
    },
    "CDT": {
        "description": "CUSTOMER DETAILS",
        "elements": [
            {
                "code": "CIDN",
                "description": "Customer's Identity",
                "components": [
                    ("Customer's EAN Location Number", "X"),
                    ("Customer's Identity Allocated by Customer", "X"),
                ],
            },
            {
                "code": "CNAM",
                "description": "Customer's Name",
                "components": [("Customer's Name", "X")],
            },
            {
                "code": "CADD",
                "description": "Customer's Address",
                "components": [
                    ("Customer's Address Line 1", "X"),
                    ("Customer's Address Line 2", "X"),
                    ("Customer's Address Line 3", "X"),
                    ("Customer's Address Line 4", "X"),
                    ("Customer's Post Code", "X"),
                ],
            },
            {
                "code": "VATR",
                "description": "Customer's VAT Registration Number",
                "components": [
                    ("Numeric VAT Registration Number", "X"),
                    ("Alphanumeric VAT Registration Number", "X"),
                ],
            },
        ],
    },
    "CDA": {
        "description": "CONTRACT DATA",
        "elements": [
            {
                "code": "CPSC",
                "description": "Current Price Structure Reference",
                "components": [
                    ("Current Price Structure Reference", "X"),
                ],
            },
            {
                "code": "ORNO",
                "description": "Order Number and Date",
                "components": [
                    ("Customer's Order Number", "X"),
                    ("Supplier's Order Number", "X"),
                    ("Date Order Placed by customer", "date"),
                    ("Date Order Received by Supplier", "date"),
                ],
            },
            {
                "code": "INSD",
                "description": "Installation Date",
                "components": [("Installation Date", "date")],
            },
            {
                "code": "REPE",
                "description": "Rental Period",
                "components": [("Rental Period", "X")],
            },
        ],
    },
    "FIL": {
        "description": "FILE DETAILS",
        "elements": [
            {
                "code": "FLGN",
                "description": "File Generation Number",
                "components": [
                    ("File Generation Number", "X"),
                ],
            },
            {
                "code": "FLVN",
                "description": "File Version Number",
                "components": [("File Version Number", "X")],
            },
            {
                "code": "FLDT",
                "description": "File Creation Date",
                "components": [
                    ("File Creation Date", "X"),
                ],
            },
            {
                "code": "FLID",
                "description": "File (Reel) Identification",
                "components": [
                    ("File (Reel) Identification", "X"),
                ],
            },
        ],
    },
    "FDT": {
        "description": "FILE PERIOD DATES",
        "elements": [
            {
                "code": "IVED",
                "description": "Invoice Period End Date",
                "components": [
                    ("Invoice Period End Date", "X"),
                ],
            },
            {
                "code": "DVED",
                "description": "Delivery Period End Date",
                "components": [("Delivery Period End Date", "X")],
            },
        ],
    },
    "REF": {
        "description": "ACCOUNT REFERENCE NUMBER",
        "elements": [
            {
                "code": "REFF",
                "description": "Account Identifier",
                "components": [
                    ("Supplier's Reference", "X"),
                    ("Customer's Reference", "X"),
                ],
            },
            {
                "code": "SCRF",
                "description": "Specification / Contract References",
                "components": [
                    ("Specification No.", "X"),
                    ("Contract No.", "X"),
                ],
            },
        ],
    },
    "MTR": {
        "description": "MESSAGE TRAILER",
        "elements": [
            {
                "code": "NOSG",
                "description": "Number of Segments in Message",
                "components": [
                    ("Number of Segments in Message", "X"),
                ],
            },
        ],
    },
    "BCD": {
        "description": "BILL CONTROL DATA",
        "elements": [
            {
                "code": "IVDT",
                "description": "Date of Invoice",
                "components": [
                    ("Date of Invoice", "date"),
                ],
            },
            {
                "code": "TXDT",
                "description": "Tax-point Date",
                "components": [
                    ("Tax-point Date", "date"),
                ],
            },
            {
                "code": "INVN",
                "description": "Invoice Number",
                "components": [
                    ("Invoice Number", "X"),
                ],
            },
            {
                "code": "PBID",
                "description": "Previous Bill Date",
                "components": [
                    ("Previous Bill Date", "date"),
                ],
            },
            {
                "code": "BIFR",
                "description": "Bill Frequency Code",
                "components": [
                    ("Bill Frequency Code", "X"),
                ],
            },
            {
                "code": "BTCD",
                "description": "Bill Type Code",
                "components": [
                    ("Bill Type Code", "X"),
                ],
            },
            {
                "code": "VDAA",
                "description": "VAT Declaration for Amended Account",
                "components": [
                    ("VAT Declaration Code", "X"),
                    ("Date of Bil Withdrawn", "date"),
                    ("VAT Total Amount on Bil Withdrawn", "X"),
                    ("Amended Invoice Reference", "X"),
                    ("Premises Reference of Previous Bill", "X"),
                ],
            },
            {
                "code": "SUMO",
                "description": "Supply Period",
                "components": [
                    ("Start Date", "date"),
                    ("End Date", "date"),
                ],
            },
            {
                "code": "CLVM",
                "description": "Calorific Value in Specified Units",
                "components": [
                    ("Calorific Value", "X"),
                    ("Unit of Measure", "X"),
                ],
            },
        ],
    },
    "CCD1": {
        "description": "CONSUMPTION / CHARGE DETAILS: consumption only",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                    ("Customer's Own Location Number", "X"),
                    ("Supplier's Identity of Customer's Location", "X"),
                ],
            },
            {
                "code": "CCDE",
                "description": "Charge Type Code",
                "components": [
                    ("Consumption / Charge Indicator", "X"),
                    ("EAN-13 Article Number", "X"),
                    ("Supplier Code", "X"),
                ],
            },
            {
                "code": "TCOD",
                "description": "Tariff Code",
                "components": [
                    ("Tariff Code", "X"),
                    ("Tariff Description", "X"),
                ],
            },
            {
                "code": "TMOD",
                "description": "Tariff Code Modifier",
                "components": [
                    ("Tariff Code Modifier 1", "X"),
                    ("Tariff Code Modifier 2", "X"),
                    ("Tariff Code Modifier 3", "X"),
                    ("Tariff Code Modifier 4", "X"),
                ],
            },
            {
                "code": "MTNR",
                "description": "Meter Number",
                "components": [
                    ("Meter Number", "X"),
                ],
            },
            {
                "code": "MLOC",
                "description": "Meter Location",
                "components": [
                    ("Meter Location", "X"),
                ],
            },
            {
                "code": "PRDT",
                "description": "Present Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "PVDT",
                "description": "Previous Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "NDRP",
                "description": "Reading Period",
                "components": [
                    ("Reading Period", "X"),
                ],
            },
            {
                "code": "PRRD",
                "description": "Reading Data",
                "components": [
                    ("Present Reading", "X"),
                    ("Type", "X"),
                    ("Pevious Reading", "X"),
                    ("Type", "X"),
                ],
            },
            {
                "code": "CONS",
                "description": "Consumption (Billing Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONB",
                "description": "Consumption (Base Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "ADJF",
                "description": "Adjustment Factor",
                "components": [
                    ("Adjustment Factor Code", "X"),
                    ("Adjustment Factor Value", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONA",
                "description": "Consumption (Adjusted Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "BPRI",
                "description": "Base Price Per Unit",
                "components": [
                    ("Base Price Per Unit", "X"),
                ],
            },
            {
                "code": "NUCT",
                "description": "Number of Units for Charge Type",
                "components": [
                    ("Units Billed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CSDT",
                "description": "Charge Start Date",
                "components": [
                    ("Charge Start Date", "date"),
                ],
            },
            {
                "code": "CEDT",
                "description": "Charge End Date",
                "components": [
                    ("Charge End Date", "date"),
                ],
            },
            {
                "code": "CPPU",
                "description": "Price per Unit",
                "components": [
                    ("Price per Unit", "X"),
                ],
            },
            {
                "code": "CTOT",
                "description": "Total Charge for Charge Type",
                "components": [
                    ("Total Charge for Charge Type", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "TSUP",
                "description": "VAT - Type of Supply",
                "components": [
                    ("VAT - Type of Supply", "X"),
                ],
            },
            {
                "code": "VATC",
                "description": "VAT Rate Category Code",
                "components": [
                    ("VAT Rate Category Code", "X"),
                ],
            },
            {
                "code": "VATP",
                "description": "VAT Rate Percentage",
                "components": [
                    ("VAT Rate Percentage", "X"),
                ],
            },
            {
                "code": "MSAD",
                "description": "Meter Sub-address",
                "components": [
                    ("Sub-address Code", "X"),
                    ("Sub-address Line", "X"),
                ],
            },
        ],
    },
    "CCD2": {
        "description": "CONSUMPTION / CHARGE DETAILS: Consumption and charge only",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                    ("Customer's Own Location Number", "X"),
                    ("Supplier's Identity of Customer's Location", "X"),
                ],
            },
            {
                "code": "CCDE",
                "description": "Charge Type Code",
                "components": [
                    ("Consumption / Charge Indicator", "X"),
                    ("EAN-13 Article Number", "X"),
                    ("Supplier Code", "X"),
                ],
            },
            {
                "code": "TCOD",
                "description": "Tariff Code",
                "components": [
                    ("Tariff Code", "X"),
                    ("Tariff Description", "X"),
                ],
            },
            {
                "code": "TMOD",
                "description": "Tariff Code Modifier",
                "components": [
                    ("Tariff Code Modifier 1", "X"),
                    ("Tariff Code Modifier 2", "X"),
                    ("Tariff Code Modifier 3", "X"),
                    ("Tariff Code Modifier 4", "X"),
                ],
            },
            {
                "code": "MTNR",
                "description": "Meter Number",
                "components": [
                    ("Meter Number", "X"),
                ],
            },
            {
                "code": "MLOC",
                "description": "Meter Location",
                "components": [
                    ("Meter Location", "X"),
                ],
            },
            {
                "code": "PRDT",
                "description": "Present Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "PVDT",
                "description": "Previous Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "NDRP",
                "description": "Reading Period",
                "components": [
                    ("Reading Period", "X"),
                ],
            },
            {
                "code": "PRRD",
                "description": "Reading Data",
                "components": [
                    ("Present Reading", "X"),
                    ("Type", "X"),
                    ("Pevious Reading", "X"),
                    ("Type", "X"),
                ],
            },
            {
                "code": "CONS",
                "description": "Consumption (Billing Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONB",
                "description": "Consumption (Base Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "ADJF",
                "description": "Adjustment Factor",
                "components": [
                    ("Adjustment Factor Code", "X"),
                    ("Adjustment Factor Value", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONA",
                "description": "Consumption (Adjusted Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "BPRI",
                "description": "Base Price Per Unit",
                "components": [
                    ("Base Price Per Unit", "X"),
                ],
            },
            {
                "code": "NUCT",
                "description": "Number of Units for Charge Type",
                "components": [
                    ("Units Billed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CSDT",
                "description": "Charge Start Date",
                "components": [
                    ("Charge Start Date", "date"),
                ],
            },
            {
                "code": "CEDT",
                "description": "Charge End Date",
                "components": [
                    ("Charge End Date", "date"),
                ],
            },
            {
                "code": "CPPU",
                "description": "Price per Unit",
                "components": [
                    ("Price per Unit", "X"),
                ],
            },
            {
                "code": "CTOT",
                "description": "Total Charge for Charge Type",
                "components": [
                    ("Total Charge for Charge Type", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "TSUP",
                "description": "VAT - Type of Supply",
                "components": [
                    ("VAT - Type of Supply", "X"),
                ],
            },
            {
                "code": "VATC",
                "description": "VAT Rate Category Code",
                "components": [
                    ("VAT Rate Category Code", "X"),
                ],
            },
            {
                "code": "VATP",
                "description": "VAT Rate Percentage",
                "components": [
                    ("VAT Rate Percentage", "X"),
                ],
            },
            {
                "code": "MSAD",
                "description": "Meter Sub-address",
                "components": [
                    ("Sub-address Code", "X"),
                    ("Sub-address Line", "X"),
                ],
            },
        ],
    },
    "CCD3": {
        "description": "CONSUMPTION / CHARGE DETAILS: Consumption based charges only",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                    ("Customer's Own Location Number", "X"),
                    ("Supplier's Identity of Customer's Location", "X"),
                ],
            },
            {
                "code": "CCDE",
                "description": "Charge Type Code",
                "components": [
                    ("Consumption / Charge Indicator", "X"),
                    ("EAN-13 Article Number", "X"),
                    ("Supplier Code", "X"),
                ],
            },
            {
                "code": "TCOD",
                "description": "Tariff Code",
                "components": [
                    ("Tariff Code", "X"),
                    ("Tariff Description", "X"),
                ],
            },
            {
                "code": "TMOD",
                "description": "Tariff Code Modifier",
                "components": [
                    ("Tariff Code Modifier 1", "X"),
                    ("Tariff Code Modifier 2", "X"),
                    ("Tariff Code Modifier 3", "X"),
                    ("Tariff Code Modifier 4", "X"),
                ],
            },
            {
                "code": "MTNR",
                "description": "Meter Number",
                "components": [
                    ("Meter Number", "X"),
                ],
            },
            {
                "code": "MLOC",
                "description": "Meter Location",
                "components": [
                    ("Meter Location", "X"),
                ],
            },
            {
                "code": "PRDT",
                "description": "Present Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "PVDT",
                "description": "Previous Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "NDRP",
                "description": "Reading Period",
                "components": [
                    ("Reading Period", "X"),
                ],
            },
            {
                "code": "PRRD",
                "description": "Reading Data",
                "components": [
                    ("Present Reading", "X"),
                    ("Type", "X"),
                    ("Pevious Reading", "X"),
                    ("Type", "X"),
                ],
            },
            {
                "code": "CONS",
                "description": "Consumption (Billing Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONB",
                "description": "Consumption (Base Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "ADJF",
                "description": "Adjustment Factor",
                "components": [
                    ("Adjustment Factor Code", "X"),
                    ("Adjustment Factor Value", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONA",
                "description": "Consumption (Adjusted Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "BPRI",
                "description": "Base Price Per Unit",
                "components": [
                    ("Base Price Per Unit", "X"),
                ],
            },
            {
                "code": "NUCT",
                "description": "Number of Units for Charge Type",
                "components": [
                    ("Units Billed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CSDT",
                "description": "Charge Start Date",
                "components": [
                    ("Charge Start Date", "date"),
                ],
            },
            {
                "code": "CEDT",
                "description": "Charge End Date",
                "components": [
                    ("Charge End Date", "date"),
                ],
            },
            {
                "code": "CPPU",
                "description": "Price per Unit",
                "components": [
                    ("Price per Unit", "X"),
                ],
            },
            {
                "code": "CTOT",
                "description": "Total Charge for Charge Type",
                "components": [
                    ("Total Charge for Charge Type", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "TSUP",
                "description": "VAT - Type of Supply",
                "components": [
                    ("VAT - Type of Supply", "X"),
                ],
            },
            {
                "code": "VATC",
                "description": "VAT Rate Category Code",
                "components": [
                    ("VAT Rate Category Code", "X"),
                ],
            },
            {
                "code": "VATP",
                "description": "VAT Rate Percentage",
                "components": [
                    ("VAT Rate Percentage", "X"),
                ],
            },
            {
                "code": "MSAD",
                "description": "Meter Sub-address",
                "components": [
                    ("Sub-address Code", "X"),
                    ("Sub-address Line", "X"),
                ],
            },
        ],
    },
    "CCD4": {
        "description": "CONSUMPTION / CHARGE DETAILS: Consumption based charges only",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                    ("Customer's Own Location Number", "X"),
                    ("Supplier's Identity of Customer's Location", "X"),
                ],
            },
            {
                "code": "CCDE",
                "description": "Charge Type Code",
                "components": [
                    ("Consumption / Charge Indicator", "X"),
                    ("EAN-13 Article Number", "X"),
                    ("Supplier Code", "X"),
                ],
            },
            {
                "code": "TCOD",
                "description": "Tariff Code",
                "components": [
                    ("Tariff Code", "X"),
                    ("Tariff Description", "X"),
                ],
            },
            {
                "code": "TMOD",
                "description": "Tariff Code Modifier",
                "components": [
                    ("Tariff Code Modifier 1", "X"),
                    ("Tariff Code Modifier 2", "X"),
                    ("Tariff Code Modifier 3", "X"),
                    ("Tariff Code Modifier 4", "X"),
                ],
            },
            {
                "code": "MTNR",
                "description": "Meter Number",
                "components": [
                    ("Meter Number", "X"),
                ],
            },
            {
                "code": "MLOC",
                "description": "Meter Location",
                "components": [
                    ("Meter Location", "X"),
                ],
            },
            {
                "code": "PRDT",
                "description": "Present Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "PVDT",
                "description": "Previous Read Date",
                "components": [
                    ("Present Read Date", "date"),
                ],
            },
            {
                "code": "NDRP",
                "description": "Reading Period",
                "components": [
                    ("Reading Period", "X"),
                ],
            },
            {
                "code": "PRRD",
                "description": "Reading Data",
                "components": [
                    ("Present Reading", "X"),
                    ("Type", "X"),
                    ("Pevious Reading", "X"),
                    ("Type", "X"),
                ],
            },
            {
                "code": "CONS",
                "description": "Consumption (Billing Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONB",
                "description": "Consumption (Base Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "ADJF",
                "description": "Adjustment Factor",
                "components": [
                    ("Adjustment Factor Code", "X"),
                    ("Adjustment Factor Value", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CONA",
                "description": "Consumption (Adjusted Units)",
                "components": [
                    ("Units Consumed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "BPRI",
                "description": "Base Price Per Unit",
                "components": [
                    ("Base Price Per Unit", "X"),
                ],
            },
            {
                "code": "NUCT",
                "description": "Number of Units for Charge Type",
                "components": [
                    ("Units Billed", "X"),
                    ("Unit of Measure", "X"),
                    ("Negative Indicator", "X"),
                ],
            },
            {
                "code": "CSDT",
                "description": "Charge Start Date",
                "components": [
                    ("Charge Start Date", "date"),
                ],
            },
            {
                "code": "CEDT",
                "description": "Charge End Date",
                "components": [
                    ("Charge End Date", "date"),
                ],
            },
            {
                "code": "CPPU",
                "description": "Price per Unit",
                "components": [
                    ("Price per Unit", "X"),
                ],
            },
            {
                "code": "CTOT",
                "description": "Total Charge for Charge Type",
                "components": [
                    ("Total Charge for Charge Type", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "TSUP",
                "description": "VAT - Type of Supply",
                "components": [
                    ("VAT - Type of Supply", "X"),
                ],
            },
            {
                "code": "VATC",
                "description": "VAT Rate Category Code",
                "components": [
                    ("VAT Rate Category Code", "X"),
                ],
            },
            {
                "code": "VATP",
                "description": "VAT Rate Percentage",
                "components": [
                    ("VAT Rate Percentage", "X"),
                ],
            },
            {
                "code": "MSAD",
                "description": "Meter Sub-address",
                "components": [
                    ("Sub-address Code", "X"),
                    ("Sub-address Line", "X"),
                ],
            },
        ],
    },
    "CLO": {
        "description": "CUSTOMER'S LOCATION",
        "elements": [
            {
                "code": "CLOC",
                "description": "Customer's Location",
                "components": [
                    ("Customer's EAN Location Number", "X"),
                    ("Customer's Own Location Number", "X"),
                    ("Supplier's Identity of Customer's Location", "X"),
                ],
            },
            {
                "code": "CNAM",
                "description": "Customer's Name",
                "components": [
                    ("Customer's Name", "X"),
                ],
            },
            {
                "code": "CADD",
                "description": "Customer's Address",
                "components": [
                    ("Customer's Address Line 1", "X"),
                    ("Customer's Address Line 2", "X"),
                    ("Customer's Address Line 3", "X"),
                    ("Customer's Address Line 4", "X"),
                    ("Customer's Post Code", "X"),
                ],
            },
        ],
    },
    "MAN": {
        "description": "METER ADMINISTRATION NUMBER",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                ],
            },
            {
                "code": "SEQB",
                "description": "Second Level Sequence Number",
                "components": [
                    ("Second Level Sequence Number", "X"),
                ],
            },
            {
                "code": "MADN",
                "description": "Distribution Identifier",
                "components": [
                    ("Distribution Identifier", "X"),
                    ("Unique Reference Number", "X"),
                    ("Check Digit", "X"),
                    ("Profile Type (class)", "X"),
                    ("Meter / time switch details", "X"),
                    ("Line Loss factor (LLF)", "X"),
                ],
            },
            {
                "code": "MTNR",
                "description": "Meter Serial Number",
                "components": [
                    ("Meter Serial Number", "X"),
                ],
            },
            {
                "code": "NDIG",
                "description": "Number of Digits",
                "components": [
                    ("Number of Digits", "X"),
                ],
            },
        ],
    },
    "VAT": {
        "description": "VALUE ADDED TAX",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                ],
            },
            {
                "code": "NDVT",
                "description": "Number of Days' VAT",
                "components": [
                    ("Number of Days' VAT", "X"),
                ],
            },
            {
                "code": "PNDP",
                "description": "Percentage Qualifying for Lower / Zero VAT Rate",
                "components": [
                    ("Percentage Qualifying for Lower / Zero VAT Rate", "X"),
                ],
            },
            {
                "code": "VATC",
                "description": "VAT Rate Category Code",
                "components": [
                    ("VAT Rate Category Code", "date"),
                ],
            },
            {
                "code": "VATP",
                "description": "VAT Rate Percentage",
                "components": [
                    ("VAT Rate Percentage", "X"),
                ],
            },
            {
                "code": "UVLA",
                "description": "Total Charge for VAT Category Before VAT",
                "components": [
                    ("Total Charge Before VAT", "X"),
                    ("Credit Line Indicator", "X"),
                ],
            },
            {
                "code": "UVTT",
                "description": "VAT Amount Payable",
                "components": [
                    ("Vat Amount Payable", "X"),
                    ("Credit Line Indicator", "X"),
                ],
            },
            {
                "code": "UCSI",
                "description": "Total Charge for VAT Category including VAT",
                "components": [
                    ("Total Charge Including VAT", "X"),
                    ("Credit Line Indicator", "X"),
                ],
            },
            {
                "code": "NRIL",
                "description": "Number of Item Lines",
                "components": [
                    ("Number of Item Lines", "X"),
                ],
            },
            {
                "code": "RFLV",
                "description": "Reason for Lower / Zero VAT Rate",
                "components": [
                    ("Reason for Lower / Zero VAT Rate", "X"),
                ],
            },
        ],
    },
    "VTS": {
        "description": "VAT RATE SUMMARY",
        "elements": [
            {
                "code": "SEQA",
                "description": "First Level Sequence Number",
                "components": [
                    ("First Level Sequence Number", "X"),
                ],
            },
            {
                "code": "VATC",
                "description": "VAT Rate Category Code",
                "components": [
                    ("VAT Rate Category Code", "date"),
                ],
            },
            {
                "code": "VATP",
                "description": "VAT Rate Percentage",
                "components": [
                    ("VAT Rate Percentage", "X"),
                ],
            },
            {
                "code": "USDI",
                "description": "File Total for VAT Category (before VAT)",
                "components": [
                    ("File Total for VAT Category", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "VTVC",
                "description": "File Total VAT for VAT Category",
                "components": [
                    ("File Total VAT for VAT Category", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
            {
                "code": "UPSI",
                "description": "File Total for VAT Category (including VAT)",
                "components": [
                    ("File Total for VAT Category (including VAT)", "X"),
                    ("Credit Indicator", "X"),
                ],
            },
        ],
    },
}



import csv
from io import StringIO

from utils import match, match_tables

from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    ReportRun,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_ecoes_comparison import _process
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_http_get(mocker, sess, client):
    mocker.patch("chellow.reports.report_ecoes_comparison.threading.Thread")
    response = client.get("/reports/ecoes_comparison")
    match(response, 303)


def test_process(mocker, sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", valid_from, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    site.insert_e_supply(
        sess,
        source,
        None,
        "Dave",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7868 6232 789",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    report_run = ReportRun.insert(
        sess,
        "ecoes_comparison",
        None,
        "ecoes_comparison",
        {},
    )

    sess.commit()
    f = StringIO()
    ecoes_lines = [
        "titles",
        ",".join(
            (
                "2278676232781",
                "address-line-1",
                "address-line-2",
                "address-line-3",
                "address-line-4",
                "address-line-5",
                "address-line-6",
                "address-line-7",
                "address-line-8",
                "address-line-9",
                "post-code",
                "supplier",
                "",
                "12/10/2022",
                "mtc",
                "12/10/2022",
                "llfc",
                "12/10/2022",
                "pc",
                "",
                "measurement-class",
                "energisation-status",
                "da",
                "dc",
                "mop",
                "12/10/2022",
                "gsp-group",
                "12/10/2022",
                "dno",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "msn",
                "12/10/2022",
                "meter-type",
                "map-id",
            ),
        ),
        ",".join(
            (
                "2278686232789",
                "address-line-1",
                "address-line-2",
                "address-line-3",
                "address-line-4",
                "address-line-5",
                "address-line-6",
                "address-line-7",
                "address-line-8",
                "address-line-9",
                "post-code",
                "supplier",
                "",
                "12/10/2022",
                "mtc",
                "12/10/2022",
                "llfc",
                "12/10/2022",
                "pc",
                "",
                "measurement-class",
                "energisation-status",
                "da",
                "dc",
                "mop",
                "12/10/2022",
                "gsp-group",
                "12/10/2022",
                "dno",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "msn",
                "12/10/2022",
                "meter-type",
                "map-id",
            ),
        ),
    ]
    exclude_mpan_cores = []
    ignore_mpan_cores_msn = []
    show_ignored = True

    _process(
        sess,
        ecoes_lines,
        exclude_mpan_cores,
        ignore_mpan_cores_msn,
        f,
        show_ignored,
        report_run,
    )
    expected = [
        [
            "mpan_core",
            "mpan_core_no_spaces",
            "ecoes_pc",
            "chellow_pc",
            "ecoes_mtc",
            "ecoes_mtc_date",
            "chellow_mtc",
            "ecoes_llfc",
            "ecoes_llfc_from",
            "chellow_llfc",
            "ecoes_ssc",
            "chellow_ssc",
            "ecoes_es",
            "chellow_es",
            "ecoes_supplier",
            "ecoes_supplier_registration_from",
            "chellow_supplier",
            "chellow_supplier_contract_name",
            "ecoes_dc",
            "chellow_dc",
            "ecoes_mop",
            "ecoes_mop_appoint_date",
            "chellow_mop",
            "ecoes_gsp_group",
            "ecoes_gsp_effective_from",
            "chellow_gsp_group",
            "ecoes_msn",
            "ecoes_msn_install_date",
            "chellow_msn",
            "ecoes_meter_type",
            "chellow_meter_type",
            "ignored",
            "problem",
        ],
        [
            "22 7867 6232 781",
            "2278676232781",
            "pc",
            "00",
            "mtc",
            "2022-10-12 00:00",
            "845",
            "llfc",
            "2022-10-12 00:00",
            "510",
            "",
            "",
            "energisation-status",
            "E",
            "supplier",
            "2022-10-12 00:00",
            "CALB",
            "Fusion Supplier 2000",
            "dc",
            "CALB",
            "mop",
            "2022-10-12 00:00",
            "CALB",
            "gsp-group",
            "2022-10-12 00:00",
            "_L",
            "msn",
            "2022-10-12 00:00",
            "hgjeyhuw",
            "meter-type",
            "H",
            "False",
            "The energisation statuses don't match. Can't parse the PC. Can't parse "
            "the MTC. The LLFCs don't match. The supplier codes don't match. The DC "
            "codes don't match. The MOP codes don't match. The GSP group codes don't "
            "match. The meter serial numbers don't match. The DTC meter types don't "
            "match.",
        ],
        [
            "22 7868 6232 789",
            "2278686232789",
            "pc",
            "00",
            "mtc",
            "2022-10-12 00:00",
            "845",
            "llfc",
            "2022-10-12 00:00",
            "510",
            "",
            "",
            "energisation-status",
            "E",
            "supplier",
            "2022-10-12 00:00",
            "CALB",
            "Fusion Supplier 2000",
            "dc",
            "CALB",
            "mop",
            "2022-10-12 00:00",
            "CALB",
            "gsp-group",
            "2022-10-12 00:00",
            "_L",
            "msn",
            "2022-10-12 00:00",
            "hgjeyhuw",
            "meter-type",
            "H",
            "False",
            "The energisation statuses don't match. Can't parse the PC. Can't parse "
            "the MTC. The LLFCs don't match. The supplier codes don't match. The DC "
            "codes don't match. The MOP codes don't match. The GSP group codes don't "
            "match. The meter serial numbers don't match. The DTC meter types don't "
            "match.",
        ],
    ]
    f.seek(0)
    actual = [line for line in csv.reader(f)]
    match_tables(expected, actual)


def test_process_in_chellow_not_ecoes(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    dno.insert_llfc(sess, "521", "Export (HV)", voltage_level, False, False, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")

    site.insert_e_supply(
        sess,
        source,
        None,
        "Dave",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7868 6232 789",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    report_run = ReportRun.insert(
        sess,
        "ecoes_comparison",
        None,
        "ecoes_comparison",
        {},
    )

    sess.commit()
    f = StringIO()
    ecoes_lines = ["titles"]
    exclude_mpan_cores = []
    ignore_mpan_cores_msn = []
    show_ignored = True

    _process(
        sess,
        ecoes_lines,
        exclude_mpan_cores,
        ignore_mpan_cores_msn,
        f,
        show_ignored,
        report_run,
    )
    expected = [
        [
            "mpan_core",
            "mpan_core_no_spaces",
            "ecoes_pc",
            "chellow_pc",
            "ecoes_mtc",
            "ecoes_mtc_date",
            "chellow_mtc",
            "ecoes_llfc",
            "ecoes_llfc_from",
            "chellow_llfc",
            "ecoes_ssc",
            "chellow_ssc",
            "ecoes_es",
            "chellow_es",
            "ecoes_supplier",
            "ecoes_supplier_registration_from",
            "chellow_supplier",
            "chellow_supplier_contract_name",
            "ecoes_dc",
            "chellow_dc",
            "ecoes_mop",
            "ecoes_mop_appoint_date",
            "chellow_mop",
            "ecoes_gsp_group",
            "ecoes_gsp_effective_from",
            "chellow_gsp_group",
            "ecoes_msn",
            "ecoes_msn_install_date",
            "chellow_msn",
            "ecoes_meter_type",
            "chellow_meter_type",
            "ignored",
            "problem",
        ],
        [
            "22 7868 6232 789",
            "2278686232789",
            "",
            "00",
            "",
            "",
            "845",
            "",
            "",
            "510",
            "",
            "",
            "",
            "E",
            "",
            "",
            "CALB",
            "Fusion Supplier 2000",
            "",
            "CALB",
            "",
            "",
            "CALB",
            "",
            "",
            "_L",
            "",
            "",
            "hgjeyhuw",
            "",
            "H",
            "False",
            "In Chellow, but not in ECOES.",
        ],
    ]
    f.seek(0)
    actual = list(csv.reader(f))
    match_tables(expected, actual)


def test_process_dtc_none(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    dno.insert_llfc(sess, "521", "Export (HV)", voltage_level, False, False, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")

    site.insert_e_supply(
        sess,
        source,
        None,
        "Dave",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        None,
        "22 7868 6232 789",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    report_run = ReportRun.insert(
        sess,
        "ecoes_comparison",
        None,
        "ecoes_comparison",
        {},
    )

    sess.commit()
    f = StringIO()
    ecoes_lines = [
        "titles",
        ",".join(
            (
                "2278686232789",
                "address-line-1",
                "address-line-2",
                "address-line-3",
                "address-line-4",
                "address-line-5",
                "address-line-6",
                "address-line-7",
                "address-line-8",
                "address-line-9",
                "post-code",
                "CALB",
                "supplier-dip",
                "12/10/2022",
                "845",
                "12/10/2022",
                "510",
                "12/10/2022",
                "00",
                "",
                "measurement-class",
                "E",
                "CALB",
                "CALB",
                "CALB",
                "12/10/2022",
                "_L",
                "12/10/2022",
                "22",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "hgjeyhuw",
                "12/10/2022",
                "",
                "map-id",
            ),
        ),
    ]
    exclude_mpan_cores = []
    ignore_mpan_cores_msn = []
    show_ignored = True

    _process(
        sess,
        ecoes_lines,
        exclude_mpan_cores,
        ignore_mpan_cores_msn,
        f,
        show_ignored,
        report_run,
    )
    expected = [
        [
            "mpan_core",
            "mpan_core_no_spaces",
            "ecoes_pc",
            "chellow_pc",
            "ecoes_mtc",
            "ecoes_mtc_date",
            "chellow_mtc",
            "ecoes_llfc",
            "ecoes_llfc_from",
            "chellow_llfc",
            "ecoes_ssc",
            "chellow_ssc",
            "ecoes_es",
            "chellow_es",
            "ecoes_supplier",
            "ecoes_supplier_registration_from",
            "chellow_supplier",
            "chellow_supplier_contract_name",
            "ecoes_dc",
            "chellow_dc",
            "ecoes_mop",
            "ecoes_mop_appoint_date",
            "chellow_mop",
            "ecoes_gsp_group",
            "ecoes_gsp_effective_from",
            "chellow_gsp_group",
            "ecoes_msn",
            "ecoes_msn_install_date",
            "chellow_msn",
            "ecoes_meter_type",
            "chellow_meter_type",
            "ignored",
            "problem",
        ],
    ]
    f.seek(0)
    actual = list(csv.reader(f))
    match_tables(expected, actual)
